from __future__ import absolute_import, print_function, unicode_literals
import mock
import pytest

from dockalot.docker import parse_args, \
    pull_base_image, make_container, run_command_list, tag_image


@pytest.fixture
def mock_docker_client():
    return mock.MagicMock()


@pytest.mark.parametrize('args,attrname,expected_value', [
    ([], 'configfile', 'fake_configfile.yml'),
    (['--ask-vault-pass'], 'ansible_args', ['--ask-vault-pass']),
    (['-e', 'v1=a', '-e', 'v2=b'], 'ansible_args',
        ['-e', 'v1=a', '-e', 'v2=b']),
    (['--vault-password-file', 'passwd.txt'], 'ansible_args',
        ['--vault-password-file', 'passwd.txt']),
])
def test_base_parse_args(args, attrname, expected_value):
    """Verify parsing of the base command line args"""
    result = parse_args(args=args + ['fake_configfile.yml'])
    assert getattr(result, attrname) == expected_value


@pytest.mark.parametrize('args,attrname,expected_value', [
    ([], 'pull', False),
    (['--label', 'a=1', '--label', 'b=2'], 'label', ['a=1', 'b=2']),
    (['--pull'], 'pull', True),
    (['-t', 't1', '-t', 't2'], 'tag', ['t1', 't2']),
])
def test_docker_parse_args(args, attrname, expected_value):
    """Verify parsing of the Docker command line args"""
    result = parse_args(args=args + ['fake_configfile.yml'])
    assert getattr(result, attrname) == expected_value


@pytest.mark.parametrize('have_image,always_pull,expect_pull', [
    (False, False, True),
    (True, False, False),
    (True, True, True),
])
def test_pull_base_image(have_image, always_pull, expect_pull,
        mock_docker_client):
    image_name = 'junkapotamus:1.0'
    config = {'always_pull': always_pull, 'docker': {'base_image': image_name}}
    mock_docker_client.images.return_value = [{'Id': 'abc'}] if have_image \
        else []
    mock_docker_client.pull.return_value = '{"status": "great"}'

    pull_base_image(config, mock_docker_client)

    if expect_pull:
        mock_docker_client.pull.assert_called_with(repository='junkapotamus',
            tag='1.0')
    else:
        mock_docker_client.pull.assert_not_called()


def test_make_container(mock_docker_client):
    image_name = 'junkapotamus:1.0'
    container_id = 'abcd'
    config = {'docker': {'base_image': image_name}}

    mock_docker_client.create_container.return_value = {
        'Id': container_id,
        'Warnings': None,
    }

    assert make_container(config, mock_docker_client) == container_id
    mock_docker_client.create_container.assert_called_with(image_name,
        command='sleep 360000')
    mock_docker_client.start.assert_called_with(resource_id=container_id)


def test_run_command_list(mock_docker_client):
    fake_container_id = 'abcd'
    fake_exec_id = 'dumbledoor'
    commands = [['echo', 'hello']]

    # Set up the mocks
    mock_docker_client.exec_create.return_value = {'Id': fake_exec_id}
    mock_docker_client.exec_start.return_value = 'hello\n'
    mock_docker_client.exec_inspect.return_value = {'ExitCode': 0}

    # Run and verify the mocks were called properly
    run_command_list(commands, mock_docker_client, fake_container_id)
    mock_docker_client.exec_create.assert_called_with(
        container=fake_container_id, cmd=commands[0])
    mock_docker_client.exec_start.assert_called_with(exec_id=fake_exec_id)
    mock_docker_client.exec_inspect.assert_called_with(exec_id=fake_exec_id)


def test_tag_image(mock_docker_client):
    config = {'docker': {'tags': ['foo:1.0', 'foo:latest']}}
    image_id = 'abcdzyxw'

    tag_image(config, mock_docker_client, image_id)
    mock_docker_client.remove_image.assert_has_calls([
        mock.call(resource_id='foo:1.0'),
        mock.call(resource_id='foo:latest'),
    ])
    mock_docker_client.tag.assert_has_calls([
        mock.call(resource_id=image_id, repository='foo', tag='1.0'),
        mock.call(resource_id=image_id, repository='foo', tag='latest'),
    ])


# vim:set ts=4 sw=4 expandtab:



from __future__ import absolute_import, print_function, unicode_literals


def pytest_addoption(parser):
    parser.addoption('--integration', action='store_true',
        help='run integration tests')


# vim:set ts=4 sw=4 expandtab:



from __future__ import absolute_import, print_function, unicode_literals
import pytest


integration = pytest.mark.skipif(
    not pytest.config.getoption('--integration'),
    reason='need --integration option to run'
)


# vim:set ts=4 sw=4 expandtab:



from __future__ import absolute_import, print_function, unicode_literals
import os
import subprocess

from .markers import integration


@integration
def test_build_basic(basic_config, image_tracker, docker_client):
    subprocess.check_call(['dockalot', basic_config])
    images = image_tracker.get_image_ids()
    assert len(images) == 1

    # Check the image metadata is correct
    image_info = docker_client.inspect_image(resource_id=images[0])
    assert image_info['RepoTags'] == []
    assert image_info['Config']['Entrypoint'] == ['/entrypoint.py']


@integration
def test_build_failure(failing_config, image_tracker):
    retry_file_name = os.path.join(os.path.dirname(failing_config),
        ".tmp-{}.retry".format('failing'))

    rc = subprocess.call(['dockalot', failing_config])
    assert rc != 0

    # No image should have been created
    images = image_tracker.get_image_ids()
    assert len(images) == 0

    # The .retry file created by Ansible should be deleted
    assert not os.path.exists(retry_file_name)


@integration
def test_build_and_tag(basic_config, image_tracker, docker_client):
    subprocess.check_call(['dockalot',
        '-t', 'testimage',
        '-t', 'testimage:1.0',
        '-t', 'testimage:1.0.5',
        basic_config])

    image_ids = image_tracker.get_image_ids()
    assert len(image_ids) == 1
    image = docker_client.inspect_image(resource_id=image_ids[0])
    image_repotags = sorted(image['RepoTags'])
    assert image_repotags == \
        ['testimage:1.0', 'testimage:1.0.5', 'testimage:latest']


@integration
def test_build_complex(complex_config, image_tracker, docker_client):
    subprocess.check_call(['dockalot', complex_config])
    images = image_tracker.get_image_ids()
    assert len(images) == 1

    # Check the image metadata is correct
    image_info = docker_client.inspect_image(resource_id=images[0])
    assert sorted(image_info['RepoTags']) == \
        ['pythonapp:1.0', 'pythonapp:latest']
    assert image_info['Config']['Entrypoint'] == ['/app/entrypoint.py']
    assert image_info['Config']['Cmd'] == ['param1', 'param2']
    assert image_info['Config']['ExposedPorts'] == {'10000/tcp': {}}
    assert image_info['Config']['Labels'] == {
        'net.primeletters.test': 'True',  # <- YAML bool converted to string
        'net.primeletters.version': '1'}
    assert image_info['Config']['WorkingDir'] == '/app'
    assert image_info['Config']['Volumes'] == {'/data': {}}


# vim:set ts=4 sw=4 expandtab:



from __future__ import absolute_import, print_function, unicode_literals
import subprocess

from .markers import integration


@integration
def test_help():
    rc = subprocess.call(['dockalot', '--help'])
    assert rc == 0


@integration
def test_version():
    rc = subprocess.call(['dockalot', '--version'])
    assert rc == 0



from __future__ import absolute_import, print_function, unicode_literals
import argparse
import docker
import json
import logging
import os
import pkg_resources
import six
import subprocess
import tempfile
from yaml.loader import SafeLoader

from dockalot.config import ArgSaverAction, Config, ConfigurationError
from dockalot.file_util import safe_delete


logger = logging.getLogger('dockalot')


def split_repo_tag(repotag):
    # Tag defaults to 'latest' if no tag is specified
    try:
        repository, tag = repotag.split(':', 1)
    except ValueError:
        repository = repotag
        tag = 'latest'
    return (repository, tag)


def escape_quotes(s):
    """
    Returns a string with backslash and double-quote characters escaped.
    """
    return s.replace('\\', '\\\\').replace('"', '\\"')


def parse_args(args=None):
    try:
        version = pkg_resources.get_distribution('dockalot').version
    except pkg_resources.DistributionNotFound:
        version = 'version unknown'

    base_parser = argparse.ArgumentParser(add_help=False)
    base_parser.add_argument('--ask-vault-pass',
        action=ArgSaverAction, nargs=0, dest='ansible_args',
        help='Prompt for the password to decrypt Ansible vaults')
    base_parser.add_argument('-e',
        action=ArgSaverAction, dest='ansible_args', metavar='EXTRA_VAR',
        help='Set additional Ansible variables as key=value')
    base_parser.add_argument('--vault-password-file',
        action=ArgSaverAction, dest='ansible_args', metavar='FILE',
        help='Specify a file with the password to decrypt an Ansible vault')
    base_parser.add_argument('--version',
        action='version', version='%(prog)s {}'.format(version))
    base_parser.add_argument('configfile',
        help='Configuration file describing how to build the image')

    parser = argparse.ArgumentParser(
        parents=[base_parser],
        description='Build a Docker image with ansible')

    parser.add_argument('--label', action='append',
        help='Add a label to the image of the format name=value. This ' +
             'option can be specified multiple times to add multiple labels.')
    parser.add_argument('--pull', action='store_true',
        help='Always pull down the latest base image')
    parser.add_argument('-t', dest='tag', action='append',
        help='A name and optional tag (in the name:tag) format. This option ' +
             'can be specified multiple times to apply multiple tags.')

    # TODO pass-thru to ansible
    #   -M --module-path
    #   -v
    return parser.parse_args(args)


def load_configuration_file(filename):
    """
    Opens the configuration file and parsers the first YAML document.
    The first document is the configuration header where our configuration
    is defined. Our configuration is validated before being returned.

    The remaining part of the YAML file is read verbatim and returned as a
    string so it can be passed untouched to the next stage (ansible).

    :returns: A tuple (dict, str) containing the dictionary of our
              configuration and a string of the remaining text in the
              configuration file.
    """
    with open(filename) as fp:
        loader = SafeLoader(fp)
        header = loader.get_data()
        if not loader.check_data():
            raise ConfigurationError("No ansible playbook defined in {}"
                .format(filename))

        # The yaml library stripped out the leading '---' so we add it back.
        playbook_text = "---"

        # yaml library signifies end of file with a NUL character so we
        # check for that.
        while playbook_text[-1] != '\0':
            text = loader.prefix(1024)
            playbook_text += text
            loader.forward(len(text))
        playbook_text = playbook_text.rstrip('\0')

    # Validate our stuff
    config = Config(header)
    return (config, playbook_text)


def pull_base_image(config, docker_client):
    """
    Pulls the base image if it's not already present. If the option to always
    pull is set then this always pulls down the latest image.
    """
    repository, tag = split_repo_tag(config['docker']['base_image'])

    if not config['always_pull']:
        # Check if we already have the base image
        images = docker_client.images(name="{}:{}".format(repository, tag))
        if len(images) > 0:
            # We already have the image
            return

    # Either we don't have the image or we were configured to always pull it
    logger.info("Pulling image %s:%s", repository, tag)
    result = docker_client.pull(repository=repository, tag=tag) \
        .rstrip().split('\n')
    final_result = json.loads(result[-1])
    if 'error' in final_result:
        raise RuntimeError(final_result['error'])
    if 'status' in final_result:
        logger.info(final_result['status'])


def make_container(config, docker_client):
    """
    Creates and starts the container that ansible will run against.
    """
    container = docker_client.create_container(
        config['docker']['base_image'],
        command='sleep 360000')

    if container['Warnings'] is not None:
        # I have never seen this set but display it anyway
        logger.warn(container['Warnings'])

    container_id = container['Id']
    logger.debug("Created container, id=%s", container_id)

    docker_client.start(resource_id=container_id)
    logger.debug("Started container")

    return container_id


def run_command_list(commands, docker_client, container_id):
    """
    Runs a list of commands in the container
    """
    for command in commands:
        logger.info("Running '%s'", command)
        result = docker_client.exec_create(container=container_id, cmd=command)
        exec_id = result['Id']
        exec_output = docker_client.exec_start(exec_id=exec_id).split('\n')
        exec_info = docker_client.exec_inspect(exec_id=exec_id)
        log_level = logging.INFO if exec_info['ExitCode'] == 0 \
            else logging.ERROR
        for line in exec_output:
            logger.log(log_level, line.rstrip())
        if exec_info['ExitCode'] != 0:
            raise RuntimeError('command failed')


def run_ansible_playbook(config_file_name, config, playbook, container_name):
    # The playbook file should be in the same directory as the original
    # configuration file so that ansible will be able to find files
    # relative to the playbook.
    tmp_playbook_file_name = os.path.join(os.path.dirname(config_file_name),
        ".tmp-{}".format(os.path.basename(config_file_name)))

    inventory_fp = None
    try:
        # Populate a temporary ansible inventory file
        inventory_fp = tempfile.NamedTemporaryFile(
            prefix='tmpinventory-', suffix='.txt')

        inventory_fp.write(container_name)
        inventory_fp.write(' ansible_connection=docker')
        inventory_fp.write(' ansible_python_interpreter="/usr/bin/env python"')
        inventory_fp.write('\n')
        # Add inventory groups
        for grp in set(config.get('inventory_groups', [])):
            inventory_fp.write("[{}]\n{}\n".format(grp, container_name))
        inventory_fp.flush()

        # Write the playbook to the temp file
        with open(tmp_playbook_file_name, 'w') as fp:
            fp.write(playbook)

        ansible_args = ['-i', inventory_fp.name] + \
            config.get('ansible_args', [])
        subprocess.check_call(['ansible-playbook'] + ansible_args +
            [tmp_playbook_file_name])
    except subprocess.CalledProcessError:
        raise RuntimeError("Ansible provisioning failed")
    finally:
        # Delete our temp files
        if inventory_fp is not None:
            inventory_fp.close()
        safe_delete(tmp_playbook_file_name)

        # Delete the .retry file if one was created
        safe_delete(os.path.splitext(tmp_playbook_file_name)[0] + '.retry')


def commit_image(config, docker_client, container_id):
    dcfg = config['docker']
    extra_commands = []

    if 'cmd' in dcfg:
        extra_commands.append("CMD {}".format(json.dumps(dcfg['cmd'])))
    if 'entrypoint' in dcfg:
        extra_commands.append("ENTRYPOINT {}".format(
            json.dumps(dcfg['entrypoint'])))
    for port in dcfg.get('expose_ports', []):
        extra_commands.append("EXPOSE {}".format(port))
    for volume in dcfg.get('volumes', []):
        extra_commands.append("VOLUME {}".format(volume))
    for k, v in six.iteritems(dcfg.get('labels', {})):
        extra_commands.append("LABEL \"{}\"=\"{}\"".format(
            escape_quotes(k), escape_quotes(str(v))))
    if 'workdir' in dcfg:
        extra_commands.append("WORKDIR {}".format(dcfg['workdir']))

    image = docker_client.commit(container_id, changes=extra_commands)
    return image['Id']


def tag_image(config, docker_client, image_id):
    repotags = config['docker'].get('tags', [])
    for repotag in repotags:
        repo, tag = split_repo_tag(repotag)
        logger.info("Tagging image %s:%s", repo, tag)

        # Remove any existing tag first. This prevents anonymous images
        # from accumulating if a tag (like 'latest') is reused
        try:
            docker_client.remove_image(resource_id=repotag)
        except docker.errors.APIError:
            # Either tag was not found or it couldn't be removed because
            # it's in use. The 'tag' command will still succed in the
            # latter case.
            pass

        docker_client.tag(resource_id=image_id, repository=repo, tag=tag)


def main():
    logging.basicConfig(level=logging.INFO, format='%(levelname)s:%(message)s')
    args = parse_args()
    try:
        config, playbook = load_configuration_file(args.configfile)
        config.merge_command_line_args(args)
    except IOError as e:
        logger.error(e.strerror)
        raise SystemExit(2)
    except Exception as e:
        logger.error(e.message)
        raise SystemExit(2)

    container_id = None
    try:
        logger.debug("Connecting to Docker daemon")
        docker_client = docker.from_env()

        pull_base_image(config, docker_client)

        logger.info("Creating the container to provision")
        container_id = make_container(config, docker_client)
        container_info = docker_client.inspect_container(container_id)
        container_name = container_info['Name'].lstrip('/')
        logger.info("Created a container named %s", container_name)

        logger.info("Preparing the container")
        run_command_list(config.get('preparation_commands', []),
            docker_client, container_id)

        logger.info("Running the ansible playbook")
        run_ansible_playbook(args.configfile, config, playbook, container_name)

        logger.info("Cleaning up the container")
        run_command_list(config.get('cleanup_commands', []),
            docker_client, container_id)

        logger.info("Committing the image")
        image_id = commit_image(config, docker_client, container_id)
        logger.info("Created %s", image_id)

        tag_image(config, docker_client, image_id)
    except KeyboardInterrupt as e:
        logger.error("Interrupted...")
        raise SystemExit(16)
    except Exception as e:
        logger.error(e.message)
        raise SystemExit(3)
    finally:
        # Delete the container
        try:
            docker_client.remove_container(resource_id=container_id,
                force=True)
        except:
            pass


# vim:set ts=4 sw=4 expandtab:



from __future__ import absolute_import, print_function, unicode_literals

import argparse
from collections import Mapping
import six


class ConfigurationError(Exception):
    pass


class ArgSaverAction(argparse.Action):
    """argparse action to save the option and its value in a list"""
    def __call__(self, parser, namespace, values, option_string=None):
        arglist = list(getattr(namespace, self.dest) or [])
        arglist.append(option_string)
        if self.nargs is not None:
            arglist.extend(values)
        else:
            arglist.append(values)
        setattr(namespace, self.dest, arglist)


def string_importer(value, name):
    if isinstance(value, six.string_types) or \
            isinstance(value, six.integer_types):
        return six.text_type(value)
    raise ConfigurationError("Configuration value '{}' is not a string"
        .format('.'.join(name)))


def integer_importer(value, name):
    try:
        return long(value)
    except:
        raise ConfigurationError("Configuration value '{}' is not an integer"
            .format('.'.join(name)))


def string_list_importer(value, name):
    if isinstance(value, list):
        return [string_importer(v, name + [str(i)])
            for i, v in enumerate(value)]
    raise ConfigurationError("Configuration value '{}' is not a list"
        .format('.'.join(name)))


def integer_list_importer(value, name):
    if isinstance(value, list):
        return [integer_importer(v, name + [str(i)])
            for i, v in enumerate(value)]
    raise ConfigurationError("Configuration value '{}' is not a list"
        .format('.'.join(name)))


def string_dict_importer(value, name):
    if isinstance(value, dict):
        return dict((six.text_type(k), string_importer(v, name + [k]))
            for k, v in six.iteritems(value))
    raise ConfigurationError("Configuration value '{}' is not a dict"
        .format('.'.join(name)))


def docker_section_importer(value, name):
    return DockerConfig(value, name)


class BaseConfigDict(Mapping):
    def __init__(self):
        self.items = {}

    def import_config_item(self, key, config_dict, importer=string_importer,
            required=False, prefix=None):
        """
        Imports the configuration from config_dict, translating it to the
        proper internal format using the given importer.
        """
        full_name = [key] if prefix is None else (prefix + [key])

        value = config_dict.get(key)
        if value is not None:
            imported_value = importer(value, name=full_name)
            self.items[key] = imported_value
        elif required:
            raise ConfigurationError("Configuration value for '{}' is missing"
                .format('.'.join(full_name)))

    def __len__(self):
        return len(self.items)

    def __getitem__(self, key):
        return self.items[key]

    def __iter__(self):
        return iter(self.items)


class DockerConfig(BaseConfigDict):
    """Docker image configuration section"""
    def __init__(self, config_dict, prefix):
        super(DockerConfig, self).__init__()

        # Required parameters
        self.import_config_item('base_image', config_dict, prefix=prefix,
            required=True)

        # Optional parameters
        self.import_config_item('cmd', config_dict,
            importer=string_list_importer, prefix=prefix)
        self.import_config_item('entrypoint', config_dict,
            importer=string_list_importer, prefix=prefix)
        self.import_config_item('expose_ports', config_dict,
            importer=integer_list_importer, prefix=prefix)
        self.import_config_item('labels', config_dict,
            importer=string_dict_importer, prefix=prefix)
        self.import_config_item('tags', config_dict,
            importer=string_list_importer, prefix=prefix)
        self.import_config_item('volumes', config_dict,
            importer=string_list_importer, prefix=prefix)
        self.import_config_item('workdir', config_dict, prefix=prefix)

    def merge_command_line_args(self, args):
        if args.tag is not None:
            self.items['tags'] = args.tag
        if args.label is not None:
            new_label_list = map(lambda l: l.split('=', 1))
            new_labels = dict((k.strip(), v.strip())
                for k, v in new_label_list)
            self.items['labels'] = new_labels


class Config(BaseConfigDict):
    """Top-level configuration dictionary"""
    def __init__(self, config_dict):
        super(Config, self).__init__()

        self.import_config_item('inventory_groups', config_dict,
            importer=string_list_importer, required=False)

        self.import_config_item('preparation_commands', config_dict,
            importer=string_list_importer, required=False)
        self.import_config_item('cleanup_commands', config_dict,
            importer=string_list_importer, required=False)

        self.import_config_item('docker', config_dict,
            importer=docker_section_importer, required=True)

    def merge_command_line_args(self, args):
        """
        Merge the values specified on the command line into the configuration.
        """
        self.items['always_pull'] = args.pull
        if args.ansible_args is not None:
            self.items['ansible_args'] = args.ansible_args
        self.items['docker'].merge_command_line_args(args)


# vim:set ts=4 sw=4 expandtab:



from __future__ import absolute_import, print_function, unicode_literals

import os


def safe_delete(filename):
    """Try to delete a file and ignore all errors"""
    try:
        os.remove(filename)
    except:
        pass


# vim:set ts=4 sw=4 expandtab:



from setuptools import find_packages, setup


setup(
    name='dockalot',
    description=('Build Docker images for the real world '
        'using ansible playbooks'),
    author='Mark Aikens',
    author_email='markadev@primeletters.net',
    license='MIT',
    url='https://github.com/markadev/dockalot',
    long_description=open('README.rst').read(),
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Environment :: Console',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Operating System :: POSIX',
        'Programming Language :: Python',
        'Topic :: System :: Installation/Setup',
        'Topic :: Utilities',
    ],

    packages=find_packages(),
    entry_points={
        'console_scripts': ['dockalot=dockalot.docker:main'],
    },

    install_requires=[
        'ansible>=2.1',
        'docker-py',
        'pyyaml',
        'six',
    ],

    tests_require=[
        'docker-py',
        'mock',
        'pytest',
    ],

    use_scm_version=True,
    setup_requires=['setuptools_scm', 'pytest-runner'],
)

# vim:set ts=4 sw=4 expandtab:



from __future__ import absolute_import, print_function, unicode_literals
import docker
import pkg_resources
import pytest


@pytest.fixture
def basic_config():
    return pkg_resources.resource_filename(__name__, 'basic.yml')


@pytest.fixture
def failing_config():
    return pkg_resources.resource_filename(__name__, 'failing.yml')


@pytest.fixture
def complex_config():
    return pkg_resources.resource_filename(__name__, 'complex.yml')


@pytest.fixture(scope='session')
def docker_client():
    return docker.from_env()


@pytest.fixture(scope='session')
def base_image(docker_client):
    """Pulls the base image used for building all the test images."""
    docker_client.pull('python', tag='2.7-slim')


class ImageTracker(object):
    def __init__(self, docker_client):
        self.docker_client = docker_client
        self.initial_imageset = set(self._image_id_list())

    def _image_id_list(self):
        return [img['Id'] for img in self.docker_client.images()]

    def get_image_ids(self):
        """Returns a list of new images created since test start"""
        current_imageset = set(self._image_id_list())
        return list(current_imageset - self.initial_imageset)

    def remove_images(self):
        """Removes all the images created since test start"""
        for img_id in self.get_image_ids():
            # Get and remove all tags for the image first
            img_info = self.docker_client.inspect_image(resource_id=img_id)
            repotags = img_info.get('RepoTags', [])
            if repotags is None:
                repotags = []
            for repotag in repotags:
                self.docker_client.remove_image(resource_id=repotag)

            # If there were no tags then remove by the image ID
            if len(repotags) == 0:
                self.docker_client.remove_image(resource_id=img_id)


@pytest.fixture
def image_tracker(base_image, docker_client):
    tracker = ImageTracker(docker_client)
    yield tracker
    tracker.remove_images()


# vim:set ts=4 sw=4 expandtab:



from __future__ import absolute_import, print_function, unicode_literals
import mock
import pytest

from dockalot.config import ConfigurationError, Config, DockerConfig, \
    string_importer, integer_importer, \
    string_list_importer, integer_list_importer, \
    string_dict_importer


@pytest.mark.parametrize('importer,input,expected_value', [
    (string_importer, 'abc', 'abc'),
    (string_importer, 123, '123'),
    (integer_importer, 123, 123),
    (integer_importer, 123L, 123),
    (integer_importer, '123', 123),
    (string_list_importer, [], []),
    (string_list_importer, [1, 2L, '3'], ['1', '2', '3']),
    (integer_list_importer, [], []),
    (integer_list_importer, [1, 2L, '3'], [1, 2, 3]),
    (string_dict_importer, {'a': 'abcd', 'b': 12}, {'a': 'abcd', 'b': '12'}),
])
def test_importer_valid(importer, input, expected_value):
    """Test low-level importers successfully convert values"""
    assert importer(input, ['test']) == expected_value


@pytest.mark.parametrize('importer,input', [
    (string_importer, ['abcd']),
    (string_importer, {'a': 1}),
    (string_importer, object()),
    (integer_importer, 'abcd'),
    (integer_importer, ['1']),
    (integer_importer, {'a': '1'}),
    (string_list_importer, 'not_a_list'),
    (string_list_importer, ['a', 'b', []]),
    (integer_list_importer, 'not_a_list'),
    (integer_list_importer, ['abcd']),
    (string_dict_importer, {'a': ['abcd', 'efgh']}),
])
def test_importer_invalid(importer, input):
    """Test low-level importers successfully reject bad values"""
    with pytest.raises(ConfigurationError):
        importer(input, ['test'])


@pytest.mark.parametrize('config_dict,key,expected_value', [
    ({}, 'docker', {}),
    ({'inventory_groups': ['a']}, 'inventory_groups', ['a']),
    ({'preparation_commands': ['xxx']}, 'preparation_commands', ['xxx']),
    ({'cleanup_commands': ['yyy']}, 'cleanup_commands', ['yyy']),
])
@mock.patch('dockalot.config.docker_section_importer', return_value={})
def test_Config_validation(m, config_dict, key, expected_value):
    """Tests successful validation of top-level configuration items"""
    # Add a fake 'docker' section because it is required
    config_dict['docker'] = {}

    cfg = Config(config_dict)
    assert cfg[key] == expected_value


@pytest.mark.parametrize('config_dict,key,expected_value', [
    ({'base_image': 'debian'},
        'base_image', 'debian'),
    ({'base_image': 'debian', 'cmd': ['bash']},
        'cmd', ['bash']),
    ({'base_image': 'debian', 'entrypoint': ['/start.sh']},
        'entrypoint', ['/start.sh']),
    ({'base_image': 'debian', 'expose_ports': [123, 345]},
        'expose_ports', [123, 345]),
    ({'base_image': 'debian', 'labels': {'a': 'wee', 'b': 2}},
        'labels', {'a': 'wee', 'b': '2'}),
    ({'base_image': 'debian', 'volumes': ['v1', 'v2']},
        'volumes', ['v1', 'v2']),
    ({'base_image': 'debian', 'workdir': '/root'},
        'workdir', '/root'),
])
def test_DockerConfig_validation(config_dict, key, expected_value):
    """
    Tests successful validation of items in the docker image configuration
    """
    c = DockerConfig(config_dict, prefix=['docker'])
    assert c[key] == expected_value


@pytest.mark.parametrize('config_dict', [
    {},  # Missing 'base_image'
    {'base_image': ['not', 'a', 'string']},
    {'base_image': 'debian', 'cmd': 'not_a_list'},
    {'base_image': 'debian', 'entrypoint': 'not_a_list'},
    {'base_image': 'debian', 'expose_ports': ['nan', 'nan']},
    {'base_image': 'debian', 'labels': 'not_a_dict'},
    {'base_image': 'debian', 'volumes': 'not_a_list'},
])
def test_DockerConfig_validation_failure(config_dict):
    with pytest.raises(ConfigurationError):
        DockerConfig(config_dict, prefix=['docker'])


# vim:set ts=4 sw=4 expandtab:
